#include "stdafx.h"
#include "Camera.h"

Camera::Camera()
{
	m_lockCamera = gcnew System::Threading::Mutex;
}

Camera::~Camera()
{
	delete m_lockCamera;
}

System::Void Camera::LockCamera(System::Void)
{
	if (m_lockCamera!= nullptr)
	{
		m_lockCamera->WaitOne();
	}
}

System::Void Camera::UnlockCamera(System::Void)
{
	if (m_lockCamera != nullptr)
	{
		m_lockCamera->ReleaseMutex();
	}
}

System::Boolean Camera::NetworkConnect(System::String^ ip_address, System::Int32 port)
{
	System::Boolean result(false);

	// Create socket
	m_networkSocket = gcnew System::Net::Sockets::TcpClient;

	if (m_networkSocket != nullptr)
	{
		try
		{
			// Connect
			m_networkSocket->Connect(ip_address, port);

			if (m_networkSocket->Connected)
			{
				m_networkStream = m_networkSocket->GetStream();

				// Read and discard welcome message from the camera
				System::Collections::Generic::List<System::String^> welcomeMessage;
				if (NetworkReceive(welcomeMessage))
				{
					result = true;
				}
			}
		}
		catch (...)
		{
			result = false;
		}
	}

	return result;
}

System::Void Camera::NetworkDisconnect(System::Void)
{
	LockCamera();

	// Send close command
	NetworkSend(L"CLOSE_CONNECTION");

	// Close network stream and socket
	m_networkStream->Close();
	m_networkSocket->Close();

	// Free memory
	delete m_networkStream;
	delete m_networkSocket;
	
	UnlockCamera();
}

System::Boolean Camera::NetworkSend(System::String^ to_send)
{
	System::Boolean result(false);
	
	if (m_networkStream != nullptr)
	{
		// Clean input buffer
		if (m_networkStream->DataAvailable)
		{
			array<System::Byte>^ temp = gcnew array<System::Byte>(4096);

			while (m_networkStream->DataAvailable)
			{
				m_networkStream->Read(temp, 0, temp->Length);
			}
		}
			
		array<System::Byte>^ data(nullptr);
		
		// Add command end token to the string
		to_send += "\r\n";

		data = System::Text::Encoding::ASCII->GetBytes(to_send);

		// Send command
		try
		{
			m_networkStream->Write(data, 0, data->Length);
			result = true;
		}
		catch (...)
		{
			result = false;
		}
		
		if (data != nullptr)
		{
			delete[] data;
			data = nullptr;
		}
	}
	
	return result;
}

System::Boolean Camera::NetworkReceive(System::Collections::Generic::List<System::String^>% received)
{
	System::Boolean result(false);
	
	if (m_networkStream != nullptr)
	{
		System::Int32 bytes(0);
		System::String^ rcv = System::String::Empty;
		array<System::Byte>^ data = gcnew array<System::Byte>(1);

		data[0] = 0;

		try
		{
			// Read from stream until the character with the ASCII value
			// 10 is received or the timeout is reached
			while (data[0] != 10)
			{
				m_networkStream->ReadTimeout = 2000;
				bytes = m_networkStream->Read(data, 0, data->Length);

				if (bytes > 0)
				{
					rcv += System::Text::Encoding::ASCII->GetString(data, 0, bytes);
				}
			}

			System::String^ temp;

			// Remove string termination tokens
			temp = rcv->TrimEnd('\0');
			temp = temp->TrimEnd('\n');
			temp = temp->TrimEnd('\r');
			
			if (temp->CompareTo("UNKNOWN CMD") == 0)
			{
				result = false;
				received.Add(L"UNKNOWN CMD");
			}
			else
			{
				// Split parameters seperator is the space sign
				array<System::String^>^ values = temp->Split(' ');

				for (System::Int32 i = 0; i < values->Length; i++)
				{
					received.Add(values[i]);
				}
				
				result = true;
			}
		}
		catch (...)
		{
			result = false;
		}
	}
	
	return result;
}

System::Boolean Camera::NetworkReceiveImage(unsigned char* buffer, int buffer_size)
{
	System::Boolean result(true);
	int result_data(0);

	if ((m_networkStream != nullptr) && (buffer != 0))
	{
		try
		{
			System::Int32 bytes(0);
			System::Int32 pos(0);
			System::String^ rcv = System::String::Empty;
			array<unsigned char>^ data = gcnew array<unsigned char>(2048);

			m_networkStream->ReadTimeout = 2000;

			int offset(0);
			bool result_is_read(false);

			// Read from stream until the complete image is transfered
			// or the timeout is reached
			do
			{
				bytes = m_networkStream->Read(data, 0, data->Length);

				if (!result_is_read)
				{
					result_is_read = true;
					result_data = data[0];

					offset = 1;
					bytes = bytes - offset;
				}
				else
				{
					offset = 0;
				}
				
				if ((pos + bytes) > buffer_size)
				{
					result = false;
				}
				else
				{
					System::Runtime::InteropServices::Marshal::Copy(data, 0 + offset, System::IntPtr(buffer + pos), bytes);
					pos += bytes;
				}
			} while ((pos < buffer_size) && (result));
		}
		catch (...)
		{
			result = false;
		}
	}
	else
	{
		result = false;
	}

	if (result_data != 0)
	{
		result = false;
	}

	return result;
}

System::Boolean Camera::Connect(System::String^ ip_address)
{
	return NetworkConnect(ip_address, 5000);
}

System::Void Camera::Disconnect(System::Void)
{
	NetworkDisconnect();
}

System::Boolean Camera::SetResolution(System::Int32 x, System::Int32 y)
{
	System::Boolean result(false);

	LockCamera();

	// Build command
	System::String^ cmd;
	cmd = "CAM_SETSENSORSIZE " + System::Convert::ToString(x) + " " + System::Convert::ToString(y);

	// Send command
	if (NetworkSend(cmd))
	{
		// Read result
		System::Collections::Generic::List<System::String^> received;
		if (NetworkReceive(received))
		{
			// Check result: This command returns 1 value
			if (received.Count == 1)
			{
				// Parse value
				System::Int32 status(0);

				status = System::Convert::ToInt32(received[0]);

				// Validate the result
				result = (status == 0 ? true : false);
			}
		}
	}

	UnlockCamera();
	
	return result;
}

System::Boolean Camera::GetResolution(System::Int32% x, System::Int32% y)
{
	System::Boolean result(false);
	
	LockCamera();

	// Send command
	if (NetworkSend(L"CAM_GETSENSORSIZE 0"))
	{
		// Read result
		System::Collections::Generic::List<System::String^> received;
		if (NetworkReceive(received))
		{
			// Check result: This command returns 3 values
			if (received.Count == 3)
			{
				// Parse values
				System::Int32 status(0);

				status = System::Convert::ToInt32(received[0]);
				x = System::Convert::ToInt32(received[1]);
				y = System::Convert::ToInt32(received[2]);

				// Validate the result
				result = (status == 0 ? true : false);
			}
		}
	}

	UnlockCamera();
	
	return result;
}


System::Boolean Camera::SetExposureTime(System::Int32 exp_in_us)
{
	System::Boolean result(false);

	LockCamera();

	// Build command
	System::String^ cmd;
	cmd = "CAM_SETSENSOREXPOSURE " + System::Convert::ToString(exp_in_us);

	// Send command
	if (NetworkSend(cmd))
	{
		// Read result
		System::Collections::Generic::List<System::String^> received;
		if (NetworkReceive(received))
		{
			// Check result: This command returns 1 value
			if (received.Count == 1)
			{
				// Parse value
				System::Int32 status(0);
					
				status = System::Convert::ToInt32(received[0]);
				
				// Validate the result
				result = (status == 0 ? true : false);
			}
		}
	}

	UnlockCamera();
	
	return result;
}

System::Boolean Camera::GetExposureTime(System::Int32% exp_in_us)
{
	System::Boolean result(false);

	LockCamera();

	// Send command
	if (NetworkSend("CAM_GETSENSOREXPOSURE"))
	{
		// Read result
		System::Collections::Generic::List<System::String^> received;
		if (NetworkReceive(received))
		{
			// Check result: This command returns 2 values
			if (received.Count == 2)
			{
				// Parse values
				System::Int32 status(0);

				status = System::Convert::ToInt32(received[0]);
				exp_in_us = System::Convert::ToInt32(received[1]);

				// Validate the result
				result = (status == 0 ? true : false);
			}
		}
	}
	
	UnlockCamera();
	
	return result;
}

System::Boolean Camera::SetFramerate(System::Int32 fps)
{
	System::Boolean result(false);

	LockCamera();

	// Build command
	System::String^ cmd;
	cmd = "CAM_SETSENSORSPEED " + System::Convert::ToString(fps);

	// Send command
	if (NetworkSend(cmd))
	{
		// Read result
		System::Collections::Generic::List<System::String^> received;
		if (NetworkReceive(received))
		{
			// Check result: This command returns 1 value
			if (received.Count == 1)
			{
				// Parse value
				System::Int32 status(0);

				status = System::Convert::ToInt32(received[0]);

				// Validate the result
				result = (status == 0 ? true : false);
			}
		}
	}

	UnlockCamera();
	
	return result;
}

System::Boolean Camera::GetFramerate(System::Int32% fps)
{
	System::Boolean result(false);
	
	LockCamera();

	// Send command
	if (NetworkSend(L"CAM_GETSENSORSPEED"))
	{
		// Read result
		System::Collections::Generic::List<System::String^> received;
		if (NetworkReceive(received))
		{
			// Check result: This command returns 2 values
			if (received.Count == 2)
			{
				// Parse values
				System::Int32 status(0);

				status = System::Convert::ToInt32(received[0]);
				fps = System::Convert::ToInt32(received[1]);

				// Validate the result
				result = (status == 0 ? true : false);
			}
		}
	}
	
	UnlockCamera();
	
	return result;
}

System::Boolean Camera::LiveStart(System::Boolean lowResolution)
{
	System::Boolean result(false);

	LockCamera();

	// Build command
	System::String^ cmd;
	cmd = "CAM_LIVESTART " + (lowResolution ? "1" : "0");

	// Send command
	if (NetworkSend(cmd))
	{
		// Read result
		System::Collections::Generic::List<System::String^> received;
		if (NetworkReceive(received))
		{
			// Check result: This command returns 1 value
			if (received.Count == 1)
			{
				// Parse value
				System::Int32 status(0);

				status = System::Convert::ToInt32(received[0]);

				// Validate the result
				result = (status == 0 ? true : false);
			}
		}
	}

	UnlockCamera();

	return result;
}

System::Boolean Camera::LiveStop(System::Void)
{
	System::Boolean result(false);

	LockCamera();

	// Send command
	if (NetworkSend(L"CAM_LIVESTOP"))
	{
		// Read result
		System::Collections::Generic::List<System::String^> received;
		if (NetworkReceive(received))
		{
			// Check result: This command returns 1 value
			if (received.Count == 1)
			{
				// Parse value
				System::Int32 status(0);

				status = System::Convert::ToInt32(received[0]);

				// Validate the result
				result = (status == 0 ? true : false);
			}
		}
	}

	UnlockCamera();
	
	return result;
}

System::Boolean Camera::GetFrame(unsigned char* buffer, System::Int32 buffer_size)
{
	System::Boolean result(false);
	
	LockCamera();

	// Send command
	if (NetworkSend(L"GET_IMAGE 1"))
	{
		// Receive image
		result = NetworkReceiveImage(buffer, buffer_size);
	}

	UnlockCamera();
	
	return result;
}

System::Boolean Camera::RecordSetup(System::Int32 numberOfFrames, System::Int32 trigger, System::Int32 triggerPos)
{
	System::Boolean result(false);

	LockCamera();

	// Build command
	System::String^ cmd;
	cmd = "CAM_FG_SETUP " + System::Convert::ToString(numberOfFrames) + " " + System::Convert::ToString(trigger) + " " + System::Convert::ToString(triggerPos);

	// Send command
	if (NetworkSend(cmd))
	{
		// Read result
		System::Collections::Generic::List<System::String^> received;
		if (NetworkReceive(received))
		{
			// Check result: This command returns 1 value
			if (received.Count == 1)
			{
				// Parse value
				System::Int32 status(0);

				status = System::Convert::ToInt32(received[0]);

				// Validate the result
				result = (status == 0 ? true : false);
			}
		}
	}

	UnlockCamera();
	
	return result;
}

System::Boolean Camera::RecordStart(System::Void)
{
	System::Boolean result(false);

	LockCamera();

	// Send command
	if (NetworkSend("CAM_FG_RECORDSTART"))
	{
		// Read result
		System::Collections::Generic::List<System::String^> received;
		if (NetworkReceive(received))
		{
			// Check result: This command returns 1 value
			if (received.Count == 1)
			{
				// Parse value
				System::Int32 status(0);

				status = System::Convert::ToInt32(received[0]);

				// Validate the result
				result = (status == 0 ? true : false);
			}
		}
	}

	UnlockCamera();

	return result;
}

System::Boolean Camera::RecordStop(System::Void)
{
	System::Boolean result(false);

	LockCamera();

	// Send command
	if (NetworkSend(L"CAM_FG_RECORDSTOP"))
	{
		// Read result
		System::Collections::Generic::List<System::String^> received;
		if (NetworkReceive(received))
		{
			// Check result: This command returns 1 value
			if (received.Count == 1)
			{
				// Parse value
				System::Int32 status(0);

				status = System::Convert::ToInt32(received[0]);

				// Validate the result
				result = (status == 0 ? true : false);
			}
		}
	}

	UnlockCamera();

	return result;
}

System::Boolean Camera::RecordAbort(System::Void)
{
	System::Boolean result(false);

	LockCamera();

	// Send command
	if (NetworkSend(L"CAM_FG_ABORTRECORD"))
	{
		// Read result
		System::Collections::Generic::List<System::String^> received;
		if (NetworkReceive(received))
		{
			// Check result: This command returns 1 value
			if (received.Count == 1)
			{
				// Parse value
				System::Int32 status(0);

				status = System::Convert::ToInt32(received[0]);

				// Validate the result
				result = (status == 0 ? true : false);
			}
		}
	}

	UnlockCamera();
	
	return result;
}

System::Boolean Camera::SoftwareTrigger(System::Void)
{
	System::Boolean result(false);

	LockCamera();

	// Send command
	if (NetworkSend(L"CAM_FG_SOFTTRIGGER"))
	{
		// Read result
		System::Collections::Generic::List<System::String^> received;
		if (NetworkReceive(received))
		{
			// Check result: This command returns 1 value
			if (received.Count == 1)
			{
				// Parse value
				System::Int32 status(0);

				status = System::Convert::ToInt32(received[0]);

				// Validate the result
				result = (status == 0 ? true : false);
			}
		}
	}

	UnlockCamera();
	
	return result;
}

System::Boolean Camera::EnableSoftwareTrigger(System::Void)
{
	System::Boolean result(false);

	LockCamera();

	// Build command
	System::String^ cmd;
	cmd = "CAM_FG_ENABLEFORCESOFTTRIGGER 1";

	// Send command
	if (NetworkSend(cmd))
	{
		// Read result
		System::Collections::Generic::List<System::String^> received;
		if (NetworkReceive(received))
		{
			// Check result: This command returns 1 value
			if (received.Count == 1)
			{
				// Parse value
				System::Int32 status(0);

				status = System::Convert::ToInt32(received[0]);

				// Validate the result
				result = (status == 0 ? true : false);
			}
		}
	}
	
	UnlockCamera();
	
	return result;
}

System::Boolean Camera::DisableSoftwareTrigger(System::Void)
{
	System::Boolean result(false);

	LockCamera();

	// Build command
	System::String^ cmd;
	cmd = "CAM_FG_ENABLEFORCESOFTTRIGGER 0";

	// Send command
	if (NetworkSend(cmd))
	{
		// Read result
		System::Collections::Generic::List<System::String^> received;
		if (NetworkReceive(received))
		{
			// Check result: This command returns 1 value
			if (received.Count == 1)
			{
				// Parse value
				System::Int32 status(0);

				status = System::Convert::ToInt32(received[0]);

				// Validate the result
				result = (status == 0 ? true : false);
			}
		}
	}

	UnlockCamera();

	return result;
}

System::Boolean Camera::GetRecordingStatus(System::Int32% recording_status)
{
	System::Boolean result(false);
	
	LockCamera();

	// Send command
	if (NetworkSend(L"CAM_FG_GETSTATUS"))
	{
		// Read result
		System::Collections::Generic::List<System::String^> received;
		if (NetworkReceive(received))
		{
			// Check result: This command returns 2 values
			if (received.Count == 2)
			{
				// Parse values
				System::Int32 status(0);

				status = System::Convert::ToInt32(received[0]);
				recording_status = System::Convert::ToInt32(received[1]);

				// Validate the result
				result = (status == 0 ? true : false);
			}
		}
	}
	
	UnlockCamera();
	
	return result;
}

System::Boolean Camera::PlayStart(System::Void)
{
	System::Boolean result(false);

	LockCamera();

	// Send command
	if (NetworkSend(L"CAM_PLAYSTART"))
	{
		// Read result
		System::Collections::Generic::List<System::String^> received;
		if (NetworkReceive(received))
		{
			// Check result: This command returns 1 value
			if (received.Count == 1)
			{
				// Parse value
				System::Int32 status(0);

				status = System::Convert::ToInt32(received[0]);
				
				// Validate the result
				result = (status == 0 ? true : false);
			}
		}
	}

	UnlockCamera();
	
	return result;
}

System::Boolean Camera::PlayStop(System::Void)
{
	System::Boolean result(false);

	LockCamera();

	// Send command
	if (NetworkSend(L"CAM_PLAYSTOP"))
	{
		// Read result
		System::Collections::Generic::List<System::String^> received;
		if (NetworkReceive(received))
		{
			// Check result: This command returns 1 value
			if (received.Count == 1)
			{
				// Parse value
				System::Int32 status(0);

				status = System::Convert::ToInt32(received[0]);

				// Validate the result
				result = (status == 0 ? true : false);
			}
		}
	}

	UnlockCamera();
	
	return result;
}

System::Boolean Camera::GetFrameMemory(unsigned char* buffer, System::Int32 buffer_size, System::Int32 frame_number)
{
	System::Boolean result(false);

	LockCamera();

	// Build command
	System::String^ cmd;
	cmd = "CAM_FG_GETFRAME " + System::Convert::ToString(frame_number) + " 0 1 1";

	// Send first command
	if (NetworkSend(cmd))
	{
		// Read result
		System::Collections::Generic::List<System::String^> received;
		if (NetworkReceive(received))
		{
			// Check result: This command returns 1 value
			if (received.Count == 1)
			{
				// Parse value
				System::Int32 status(0);

				status = System::Convert::ToInt32(received[0]);

				// Validate the result
				result = (status == 0 ? true : false);
			}

			if (result)
			{
				// Send second command
				result = NetworkSend("GET_IMAGE 1");
			}

			if (result)
			{
				// Read image
				result = NetworkReceiveImage(buffer, buffer_size);
			}
		}
	}
	
	UnlockCamera();
	
	return result;
}

System::Boolean Camera::GetFrameMemory(unsigned char* buffer, System::Int32 buffer_size, System::UInt64% time_marker, System::Int32 frame_number)
{
	System::Boolean result(false);

	LockCamera();

	// Build command
	System::String^ cmd;
	cmd = "CAM_FG_GETFRAME_TM " + System::Convert::ToString(frame_number) + " 0 1 1";

	// Send first command
	if (NetworkSend(cmd))
	{
		// Read result
		System::Collections::Generic::List<System::String^> received;
		if (NetworkReceive(received))
		{
			// Check result: This command returns 2 values
			if (received.Count == 2)
			{
				// Parse values
				System::Int32 status(0);

				status = System::Convert::ToInt32(received[0]);
				time_marker = System::Convert::ToUInt64(received[1]);

				// Validate the result
				result = (result == 0 ? true : false);
			}

			if (result)
			{
				// Send second command
				result = NetworkSend("GET_IMAGE 1");
			}

			if (result)
			{
				// Read image
				result = NetworkReceiveImage(buffer, buffer_size);
			}
		}
	}
	
	UnlockCamera();

	return result;
}

System::Boolean Camera::GetSequenceParameter(System::Int32% frames, 
	                                         System::Int32% resoX, 
											 System::Int32% resoY, 
											 System::Int32% fps, 
											 System::Int32% exptime, 
											 System::Int32% trigpos, 
											 System::Boolean% bFG8bits, 
											 System::Int32% mode, 
											 System::Int32% nbSeq, 
											 System::Double% realfps)
{
	System::Boolean result(false);

	LockCamera();

	// Send command
	if (NetworkSend(L"CAM_FG_GETPARAM"))
	{
		// Read result
		System::Collections::Generic::List<System::String^> received;
		if (NetworkReceive(received))
		{
			// Check result: This command returns 11 values
			if (received.Count == 11)
			{
				// Parse values
				System::Int32 status(0);

				status = System::Convert::ToInt32(received[0]);

				// Validate the result
				result = (status == 0 ? true : false);
								
				if (result)
				{
					frames		= System::Convert::ToInt32(received[1]);
					resoX		= System::Convert::ToInt32(received[2]);
					resoY		= System::Convert::ToInt32(received[3]);
					fps			= System::Convert::ToInt32(received[4]);
					exptime		= System::Convert::ToInt32(received[5]);
					trigpos		= System::Convert::ToInt32(received[6]);
					bFG8bits	= (System::Convert::ToInt32(received[7]) == 1 ? true : false);
					mode		= System::Convert::ToInt32(received[8]);
					nbSeq		= System::Convert::ToInt32(received[9]);
					realfps		= System::Convert::ToDouble(received[10]);
				}
			}
		}
	}
	
	UnlockCamera();
	
	return result;
}

System::Boolean Camera::IsSequenceAvailable(System::Boolean% available)
{
	System::Boolean result(false);

	LockCamera();

	// Build command
	System::String^ cmd;
	cmd = "CAM_FG_ISSEQUENCETHERE -1";

	// Send command
	if (NetworkSend(cmd))
	{
		// Read result
		System::Collections::Generic::List<System::String^> received;
		if (NetworkReceive(received))
		{
			// Check result: This command returns 2 values
			if (received.Count == 2)
			{
				// Parse values
				System::Int32 status(0);

				status = System::Convert::ToInt32(received[0]);
				available = (System::Convert::ToInt32(received[1]) == 0 ? true : false);

				// Validate the result
				result = (status == 0 ? true : false);
			}
		}

		UnlockCamera();
	}

	return result;
}

System::Boolean Camera::GetSensorScale(System::Int32% x_min, System::Int32% x_max, System::Int32% x_inc, System::Int32% y_min, System::Int32% y_max, System::Int32% y_inc)
{
	System::Boolean result(false);
	
	LockCamera();

	// Build command
	System::String^ cmd;
	cmd = "CAM_GETSENSORSCALE";
	
	// Send command
	if (NetworkSend(cmd))
	{
		// Read result
		System::Collections::Generic::List<System::String^> received;
		if (NetworkReceive(received))
		{
			// Check result: This command returns 7 values
			if (received.Count == 7)
			{
				// Parse values
				System::Int32 status = System::Convert::ToInt32(received[0]);

				x_min = System::Convert::ToInt32(received[1]);
				x_max = System::Convert::ToInt32(received[2]);
				x_inc = System::Convert::ToInt32(received[3]);
				y_min = System::Convert::ToInt32(received[4]);
				y_max = System::Convert::ToInt32(received[5]);
				y_inc = System::Convert::ToInt32(received[6]);

				// Validate the result
				result = (status == 0 ? true : false);
			}
		}
	}
	
	UnlockCamera();
	
	return result;
}